/********************
* PROCESS FUNCTIONS *
********************/
PROCESS_CONTEXT_ENTRY *CreateProcess(void *header, void *entrypoint)
{
  PrositHeaderElement_Stacksize *stacksizefield;
  PrositHeaderElement_Eventbuffersize *eventbuffersizefield;
  ULONG stacksize, eventbuffersize;
  PROCESS_CONTEXT_ENTRY *Context;

  // Get the wanted stacksize of the Prosit header. Use default if none is found.
  if(stacksizefield = FindPrositHeaderField(header,PROSIT_STACKSIZE))
    stacksize = stacksizefield->stacksize;
  else
    stacksize = STACK_HEADERSIZE + DEFAULT_STACKSIZE;

  // Get the wanted event buffer size of the Prosit header. Use default if none is found.
  if(eventbuffersizefield = FindPrositHeaderField(header,PROSIT_EVENTBUFFERSIZE))
    eventbuffersize = eventbuffersizefield->eventbuffersize;
  else
    eventbuffersize = DEFAULT_EVENTBUFFERSIZE;

  // Check if the PROSIT_VERSION is correct and thet the PROSIT_FILETYPE is "prog", plus
  // that there is enough space on the heap for the stack and the event buffer.
  if((((PrositHeaderElement_Version *)(FindPrositHeaderField(header,PROSIT_VERSION)))->version == PROSIT_VERSION_NUMBER) &&
     (((PrositHeaderElement_Filetype *)(FindPrositHeaderField(header,PROSIT_FILETYPE)))->filetype == PROSIT_FILETYPE_PROGRAM) &&
     (HeapAvail() > (sizeof(PROCESS_CONTEXT_ENTRY) + stacksize + eventbuffersize + EVENTBUFFER_HEADERSIZE))) {

    WORD ProcessID;
    MEMORY_LONG *stack1;
    MEMORY_WORD *stack2;

    // Allocate memory for the context.
    Context = HeapDeref(ProcessID = HeapAlloc(sizeof(PROCESS_CONTEXT_ENTRY)));
    Context->header.id = ProcessID;

    // Initialize a bunch of context stuff.
    Context->sr = 0;
    Context->pc = entrypoint;
    Context->prositheader = header;
    asm(" movem.l	(0x00007464),%%a4-%%a5/%%d0/%%d2-%%d7
          movem.l	%%a4-%%a5/%%d0/%%d2-%%d7,8+16*4+6(%0)"
      : : "a" (Context) : "a4", "a5", "d0", "d2", "d3", "d4", "d5", "d6", "d7");

    // Allocate and initialize the event buffer.
    Context->eventbufferhandle = HeapAlloc(eventbuffersize + EVENTBUFFER_HEADERSIZE);
    RBufferInitialize(HeapDeref(Context->eventbufferhandle), eventbuffersize);

    // Allocate and initialize the stack.
    if(stacksize) {
      Context->usp = HeapDeref(Context->stackhandle = HeapAlloc(stacksizefield->stacksize)) + stacksizefield->stacksize-10;
      *((ULONG *)(Context->usp)) = (ULONG)TerminateProcess;
      *((WORD *)((void *)Context->usp + 8)) = Context->header.id;
    }
    else {
      Context->stackhandle = 0;
    }

    // Set the new process' parent.
    Context->parent = CurrentProcessID;

    // Insert the new context into the context list.
    Context->header.priority = 0;
    ListInsert(&ContextList,&Context->header);
  }
  else
    Context = NULL;

  return(Context);
}

//
// Terminate a process... this is also where a self-terminated
// process will end up.
//
void TerminateProcess(HANDLE pid)
{
  PROCESS_CONTEXT_ENTRY *Context, *next;
  ListHeader *list = &ContextList;

  ToggleScheduler();

  if(!(Context = (PROCESS_CONTEXT_ENTRY *)ListFind(&ContextList, pid)))
    Context = (PROCESS_CONTEXT_ENTRY *)ListFind((list = &SleepingContextList), pid);

  if(Context != NULL) {
    //Context = HeapDeref(pid);
    if(!Context->flags.bits.fork)
      UnrelocateFile(Context->filehandle);

    next = Scheduler();
    ListRemove(list,&Context->header);
    HeapFree(Context->eventbufferhandle);
    HeapFree(Context->stackhandle);
    HeapFree(pid);

    // If the calling process is the same as the terminated process we must
    // jump into the scheduler.
    if(pid == CurrentProcessID) {
      asm("	move.l %0,%%a0
		trap #15
		bra PostScheduler" : : "a" (next));
    }
  }

  ToggleScheduler();
}

//
// Create a process and add it to the ContextList.
// Returns the forks ID, NULL if not error.
//
HANDLE NewProcess(HANDLE filehandle)
{
  PROCESS_CONTEXT_ENTRY *Context;
  void *header;
  PROSIT_FILE_HEADER *FileStart;

  ToggleScheduler();

  if((header = GetPrositHeaderOfFile(filehandle)) &&
     (FileStart = RelocateFile(filehandle)) &&
     (Context = CreateProcess(header, (void *)FileStart+FileStart->offset_main+2))) {
    Context->flags.all = CS_INIT_NORMAL;
    Context->filehandle = filehandle;

    ToggleScheduler();
    return(Context->header.id);
  }
  else {
    ToggleScheduler();
    return(NULL);
  }
}

//
// Create a fork and add it to the ContextList.
// Returns the forks ID, NULL if not error.
//
HANDLE ForkProcess(void *header, void *entrypoint)
{
  PROCESS_CONTEXT_ENTRY *Context;

  ToggleScheduler();
  if((Context = CreateProcess(header, entrypoint))) {
    Context->flags.all = (CurrentProcessContext->flags.all & CS_FORK_KEEP_FLAGS) | CS_INIT_FORK;
    Context->filehandle = CurrentProcessContext->filehandle;
  }

  ToggleScheduler();
  return(Context->header.id);
}


//
// Return number of running processes
//
int RunningProcesses(void)
{
  int ret;
  ToggleScheduler();
  ret = ListSize(&ContextList);
  ToggleScheduler();
  return ret;
}


//
// Return number of sleeping processes
//
int SleepingProcesses(void)
{
  int ret;
  ToggleScheduler();
  ret = ListSize(&SleepingContextList);
  ToggleScheduler();
  return ret;
}

//
// Move a process from one list to another
// If it's the currently running process, save it's data and give
// control to the next process that should run.
//
void MoveProcess(PROCESS_CONTEXT_ENTRY *Context, ListHeader *source, ListHeader *destination)
{
  PROCESS_CONTEXT_ENTRY *next;

  next = Scheduler();
  ListRemove(source, &Context->header);
  ListInsert(destination, &Context->header);

  if(Context == CurrentProcessContext) {
    asm("	trap		#15
		/* or.w		#0x0700,%%sr */					/* Turn off all interrupts */
		move.l		%%a6,-(%%sp)					/* Store a6 on the stack */
		move.l		CurrentProcessContext,%%a6
		movem.l		%%d0-%%d7/%%a0-%%a5,8(%%a6)			/* Store registers */
		move.l		(%%sp)+,8+14*4(%%a6)				/* Store register a6 */
		clr.w		8+16*4(%%a6)					/* Set process' SR to zero */
		move.l		#ProcessWokeUp,8+16*4+2(%%a6)			/* Set process' PC */
		move.l		%%usp,%%a1					/* Get sp (a7) */
		move.l		%%a1,8+15*4(%%a6)				/* Set process' SP */
		movem.l		(" XSTR(TIOS_CONSTANTS) "),%%a0/%%d0-%%d7
		movem.l		%%a0/%%d0-%%d7,8+16*4+6(%%a6)			/* Store process' TIOS constants */
		move.l		%0,%%a0
		bra		PostScheduler" : : "a" (next));
    asm(" ProcessWokeUp:");
  }
}

// Put a process to sleep, i.e. move it from ContextList
// to SleepingContextList.
/*
void Sleep(HANDLE pid)
{
  RBUFFER *buffer = HeapDeref(CurrentProcessContext->eventbufferhandle);
  if(!(buffer->pending)) {
    ToggleScheduler();
    CurrentProcessContext->flags.bits.sleeping = 1;
    MoveProcess(CurrentProcessContext,&ContextList,&SleepingContextList);
  }

  PROCESS_CONTEXT_ENTRY *Context, *next;

  if((Context = (PROCESS_CONTEXT_ENTRY *)ListFind(&ContextList, pid))) {
    RBUFFER *buffer = HeapDeref(CurrentProcessContext->eventbufferhandle);
    if(!(buffer->pending)) {
      ToggleScheduler();

//    asm(" move.l #0x00000000,(0x4c00)");

      next = Scheduler();
      ListRemove(&ContextList, &Context->header);
      ListInsert(&SleepingContextList, &Context->header);

      if(pid == CurrentProcessID) {
*/
//        asm("	or.w		#0x0700,%%sr					/* Turn off all interrupts */
//		move.l		%%a6,-(%%sp)					/* Store a6 on the stack */
//		move.l		CurrentProcessContext,%%a6
//		movem.l		%%d0-%%d7/%%a0-%%a5,8(%%a6)			/* Store registers */
//		move.l		(%%sp)+,8+14*4(%%a6)				/* Store register a6 */
//		clr.w		8+16*4(%%a6)					/* Set process' SR to zero */
//		move.l		#ProcessWokeUp,8+16*4+2(%%a6)			/* Set process' PC */
//		move.l		%%usp,%%a1					/* Get sp (a7) */
//		move.l		%%a1,8+15*4(%%a6)				/* Set process' SP */
//		movem.l		(" XSTR(TIOS_CONSTANTS) "),%%a0/%%d0-%%d7
//		movem.l		%%a0/%%d0-%%d7,8+16*4+6(%%a6)			/* Store process' TIOS constants */
//		move.l		%0,%%a0
//		bra		PostScheduler" : : "a" (next));
/*
      }

      ToggleScheduler();
      asm(" ProcessWokeUp:");
    }
  }
}
*/




// Wake-up a sleeping process.
/*
void WakeUp(HANDLE pid)
{
  if(pid==CurrentProcessID) {
    CurrentProcessContext->flags.bits.sleeping = 0;
  }
  ToggleScheduler();
  MoveProcess((PROCESS_CONTEXT_ENTRY *)HeapDeref(pid),&SleepingContextList,&ContextList);
  ToggleScheduler();

  PROCESS_CONTEXT_ENTRY *Context;

  if((Context = (PROCESS_CONTEXT_ENTRY *)ListFind(&SleepingContextList, pid))) {
    ToggleScheduler();

    ListRemove(&SleepingContextList, &Context->header);
    ListInsert(&ContextList, &Context->header);

    ToggleScheduler();
  }
  else if(pid = CurrentProcessID) {
    Context->flags.bits.sleeping = 0;
  }
}

*/
